from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Union

import anyio
import pandas as pd

if TYPE_CHECKING:
    from ._types import TypeAlias

FilePathType: TypeAlias = Union[str, Path, anyio.Path]


async def read_excel(file: FilePathType | bytes, as_str: bool = False, **kw) -> pd.DataFrame:
    """Read excel from local file or bytes

    :param as_str: whether to read as dtype=str
    :param kw: other kwargs that will pass to the `pd.read_excel` function
    """
    if isinstance(file, anyio.Path):
        file = await file.read_bytes()
    if as_str and "dtype" not in kw:
        kw.setdefault("dtype", str)
    kw.setdefault("keep_default_na", False)
    if isinstance(file, bytes):
        return pd.read_excel(BytesIO(file), **kw)
    else:
        return pd.read_excel(file, **kw)


def df_to_datas(df: pd.DataFrame) -> list[dict]:
    """Convert dataframe to list of dict"""
    cols = list(df.columns)
    return [dict(zip(cols, v)) for v in df.values.tolist()]


async def load_xls(file: FilePathType | bytes, as_str=False, **kw) -> list[dict]:
    """Read excel file or content to be list of dict"""
    return df_to_datas(await read_excel(file, as_str, **kw))



from collections.abc import Generator
from pathlib import Path

import pytest

from asynctor.compat import chdir


@pytest.fixture(scope="session")
def anyio_backend():
    return "asyncio"


@pytest.fixture
def tmp_work_dir(tmp_path: Path) -> Generator[Path]:
    with chdir(tmp_path):
        yield tmp_path



import random
import time

from asynctor.tasks import ThreadGroup


def sleep(seconds=0.1):
    time.sleep(seconds)
    return seconds


def test_thread_group():
    total = 10
    start = time.time()
    with ThreadGroup() as tg:
        for _ in range(total):
            tg.soonify(sleep)()
    end = time.time()
    assert round(end - start, 1) == 0.1
    assert tg.results == [0.1] * total
    start = time.time()
    with ThreadGroup() as tg:
        for _ in range(total):
            tg.soonify(sleep)(0.2)
    end = time.time()
    assert round(end - start, 1) == 0.2
    assert tg.results == [0.2] * total
    start = time.time()
    choices = [0.1, 0.2, 0.3]
    seconds = [random.choice(choices) for _ in range(total)]
    with ThreadGroup() as tg:
        for t in seconds:
            tg.soonify(sleep)(seconds=t)
    end = time.time()
    assert round(end - start, 1) == max(seconds)
    assert len(tg.results) == total
    assert all(i in choices for i in tg.results)


def test_thread_group_timeout():
    def foo():
        return 1

    start = time.time()
    with ThreadGroup(timeout=0.1) as tg:
        tg.soonify(sleep)(seconds=0.2)
        tg.soonify(foo)()
    end = time.time()
    assert round(end - start, 1) == 0.1
    assert isinstance(tg.results[0], TimeoutError)
    assert tg.results[1] == 1
    assert len(tg.results) == 2


def test_thread_group_exception():
    def raise_it(e: Exception):
        raise e

    with ThreadGroup() as tg:
        tg.soonify(raise_it)(ValueError("foo"))
        tg.soonify(raise_it)(TypeError("type"))

    assert len(tg.results) == 2
    assert isinstance(tg.results[0], ValueError)
    assert isinstance(tg.results[1], TypeError)
    assert str(tg.results[0]) == "foo"
    assert str(tg.results[1]) == "type"

    with ThreadGroup(max_workers=None) as tg:
        tg.soonify(raise_it)(ValueError("foo"))
        tg.soonify(raise_it)(TypeError("type"))

    assert len(tg.results) == 2
    assert isinstance(tg.results[0], ValueError)
    assert isinstance(tg.results[1], TypeError)
    assert str(tg.results[0]) == "foo"
    assert str(tg.results[1]) == "type"


def test_thread_group_with_pool():
    start = time.time()
    with ThreadGroup(max_workers=1) as tg:
        tg.soonify(sleep)(seconds=0.1)
        tg.soonify(sleep)(seconds=0.2)
    end = time.time()
    assert round(end - start, 1) == 0.3
    assert tg.results[0] == 0.1
    assert tg.results[1] == 0.2



from __future__ import annotations

import time
from contextlib import contextmanager, redirect_stdout
from datetime import datetime, timedelta
from io import StringIO

import anyio
import pytest

from asynctor.timing import Timer, timeit


@contextmanager
def capture_stdout():
    """Redirect sys.stdout to a new StringIO

    Example::
    ```py
        with capture_stdout() as stream:
            GitTag(message="", dry=True).run()
        assert "git tag -a" in stream.getvalue()
    ```
    """
    stream = StringIO()
    with redirect_stdout(stream):
        yield stream


@timeit
async def sleep(seconds: float | int) -> None:
    return await raw_sleep(seconds)


@Timer
async def do_sleep(seconds):
    return await raw_sleep(seconds)


async def raw_sleep(seconds):
    await anyio.sleep(seconds)


@timeit
async def sleep1() -> int:
    return await raw_sleep1()


@Timer
async def do_sleep1() -> int:
    return await raw_sleep1()


async def raw_sleep1() -> int:
    await anyio.sleep(0.1)
    return 1


@timeit
def wait_for() -> str:
    return raw_wait_for()


@Timer
def do_wait_for() -> str:
    return raw_wait_for()


def raw_wait_for():
    time.sleep(0.12)
    return "I'm a teapot"


@pytest.mark.anyio
async def test_timeit():
    start = time.time()
    s = 0.2
    with capture_stdout() as stream:
        r = await sleep(s)
    end = time.time()
    assert round(end - start, 1) == s
    assert r is None
    stdout = stream.getvalue()
    assert str(s) in stdout and sleep.__name__ in stdout
    start = time.time()
    with capture_stdout() as stream1:
        r1 = await sleep1()
    end = time.time()
    assert round(end - start, 1) == 0.1
    assert r1 == 1
    stdout1 = stream1.getvalue()
    assert "0.1" in stdout1 and sleep1.__name__ in stdout1
    start = time.time()
    with capture_stdout() as stream2:
        r2 = wait_for()
    end = time.time()
    assert round(end - start, 1) == 0.1
    assert r2 == "I'm a teapot"
    stdout2 = stream2.getvalue()
    assert "0.1" in stdout2 and wait_for.__name__ in stdout2
    assert "0.12" not in stdout2


class TestTimer:
    @pytest.mark.anyio
    async def test_decorator(self):
        start = time.time()
        s = 0.2
        with capture_stdout() as stream:
            r = await do_sleep(s)
        end = time.time()
        assert round(end - start, 1) == s
        assert r is None
        stdout = stream.getvalue()
        assert str(s) in stdout and do_sleep.__name__ in stdout
        start = time.time()
        with capture_stdout() as stream1:
            r1 = await do_sleep1()
        end = time.time()
        assert round(end - start, 1) == 0.1
        assert r1 == 1
        stdout1 = stream1.getvalue()
        assert "0.1" in stdout1 and do_sleep1.__name__ in stdout1
        start = time.time()
        with capture_stdout() as stream2:
            r2 = do_wait_for()
        end = time.time()
        assert round(end - start, 1) == 0.1
        assert r2 == "I'm a teapot"
        stdout2 = stream2.getvalue()
        assert "0.1" in stdout2 and do_wait_for.__name__ in stdout2
        assert "0.12" not in stdout2

    @pytest.mark.anyio
    async def test_start_capture(self):
        # Manual capture
        clock = Timer("testing start capture", decimal_places=2, verbose=False)
        assert repr(clock) == "Timer('testing start capture', 2, False)"
        start = time.time()
        assert clock._start <= start
        time.sleep(0.5)
        clock.start()
        assert clock._start > start
        with capture_stdout() as stream:
            await raw_sleep(0.23)
            clock.capture()
        assert not stream.getvalue()
        assert clock.cost == 0.23 or clock.cost == 0.24
        assert str(clock) == f"testing start capture Cost: {clock.cost} seconds"
        # with syntax
        with capture_stdout() as stream, Timer("Let it move:", verbose=False) as watch:
            time.sleep(0.3)
        assert not stream.getvalue()
        assert watch.cost == 0.3
        assert str(watch) == "Let it move: Cost: 0.3 seconds"
        # Explicit set verbose as True
        with capture_stdout() as stream, Timer("Let it move:", verbose=True):
            time.sleep(0.3)
        assert stream.getvalue().strip() == "Let it move: Cost: 0.3 seconds"
        # Initial with verbose False, but capture with verbose True
        pendulum = Timer("Initial verbose False but capture True", verbose=False)
        time.sleep(0.2)
        with capture_stdout() as stream:
            pendulum.capture(verbose=True)
        assert (
            stream.getvalue().strip() == "Initial verbose False but capture True Cost: 0.2 seconds"
        )
        # Initial with verbose True, capture will auto print cost message
        timepiece = Timer("I'm a teapot --")
        time.sleep(0.1)
        with capture_stdout() as stream:
            timepiece.capture()
        assert stream.getvalue().strip() == "I'm a teapot -- Cost: 0.1 seconds"

    @pytest.mark.anyio
    async def test_with(self):
        start = time.time()
        message = "Welcome to guangdong"
        with capture_stdout() as stream, Timer(message):
            await raw_sleep1()
            raw_wait_for()
            await raw_sleep(0.21)
        end = time.time()
        assert round(end - start, 1) == (0.1 + 0.1 + 0.2)
        stdout = stream.getvalue()
        assert "0.4" in stdout
        assert raw_sleep.__name__ not in stdout
        assert raw_wait_for.__name__ not in stdout
        assert raw_sleep1.__name__ not in stdout
        assert message in stdout

    def test_invalid_use_case(self):
        assert Timer("")() is None


@pytest.mark.anyio
async def test_with_timeit():
    start = time.time()
    message = "hello kitty"
    with capture_stdout() as stream, timeit(message):
        await raw_sleep1()
        raw_wait_for()
        await raw_sleep(0.21)
    end = time.time()
    assert round(end - start, 1) == (0.1 + 0.1 + 0.2)
    stdout = stream.getvalue()
    assert "0.4" in stdout
    assert raw_sleep.__name__ not in stdout
    assert raw_wait_for.__name__ not in stdout
    assert raw_sleep1.__name__ not in stdout
    assert message in stdout


def test_nows():
    utc_now = Timer.now()
    bj_now = Timer.beijing_now()
    assert str(utc_now).endswith("+00:00")
    assert str(bj_now).endswith("+08:00")
    utc_ts = utc_now.timestamp()
    bj_ts = bj_now.timestamp()
    assert bj_ts - utc_ts < 1
    assert round(utc_ts) == round(bj_ts)
    assert utc_ts == Timer.to_beijing(utc_now).timestamp()
    assert str(Timer.to_beijing(utc_now)).endswith("+08:00")
    fmt = "%Y-%m-%d %H:%M:%S"
    assert (
        datetime.strptime(str(Timer.to_beijing(utc_now)).split(".")[0], fmt)
        - datetime.strptime(str(utc_now).split(".")[0], fmt)
    ) == timedelta(hours=8)



# mypy: disable-error-code="attr-defined"
from __future__ import annotations

import re
import shlex
from datetime import datetime
from pathlib import Path

import pytest

from asynctor import AttrDict
from asynctor.utils import AsyncTestClient, Shell, cache_attr, client_manager, get_machine_ip

from .main import app_default_to_mount_lifespan, app_for_utils_test


@pytest.fixture(scope="session")
async def client():
    async with AsyncTestClient(app_default_to_mount_lifespan) as c:
        yield c


@pytest.fixture(scope="session")
async def client_without_lifespan():
    async with AsyncTestClient(app_for_utils_test, mount_lifespan=False) as c:
        yield c


@pytest.fixture(scope="session")
async def client_func_style():
    async with client_manager(app_default_to_mount_lifespan) as c:
        yield c


@pytest.fixture(scope="session")
async def client_func_style_without_lifespan():
    async with client_manager(app_for_utils_test, mount_lifespan=False) as c:
        yield c


@pytest.mark.anyio
async def test_async_test_client(client, client_func_style):
    path = "/state"
    for c in (client, client_func_style):
        r = await c.get(path)
        assert r.status_code == 200
        assert r.json()["redis"] != "None"


@pytest.mark.anyio
async def test_async_test_client2(client_without_lifespan, client_func_style_without_lifespan):
    path = "/state"
    for c in (client_without_lifespan, client_func_style_without_lifespan):
        r = await c.get(path)
        assert r.status_code == 200
        assert r.json() == {"redis": "None"}


class TestAttrDict:
    def test_normal_case(self):
        origin_dict = {"a": 1, "b": {"c": 2, "d": {"e": 3}}}
        d = AttrDict(origin_dict)
        assert d == origin_dict
        assert d.a == d["a"] == 1
        assert d.b == d["b"] == {"c": 2, "d": {"e": 3}}
        assert d.b.c == d["b"]["c"] == 2
        assert d.b.d == d["b"]["d"] == {"e": 3}
        assert d.b.d.e == d["b"]["d"]["e"] == 3
        assert str(d) == str(origin_dict)
        assert repr(d) == "AttrDict(" + repr(origin_dict) + ")"

    def test_raises(self):
        with pytest.raises(AttributeError):
            assert AttrDict().a
        with pytest.raises(KeyError):
            assert AttrDict()["a"]

    def test_key_startswith_underline(self):
        d = AttrDict({"_a": 1, "__b": 2, "___c": 3, "____d": 4})
        assert d._a == d["_a"] == 1
        assert d["__b"] == 2
        assert d["___c"] == 3
        assert d["____d"] == 4
        with pytest.raises(AttributeError):
            assert d.__b
        with pytest.raises(AttributeError):
            assert d.___c
        with pytest.raises(AttributeError):
            assert d.____d

    def test_initial(self):
        assert AttrDict() == {} == {}
        assert AttrDict(a=1) == {"a": 1} == {"a": 1}
        assert AttrDict(a=1, b=2) == {"a": 1, "b": 2} == {"a": 1, "b": 2}
        assert (
            AttrDict({1: 2}, a=1, b=2) == {1: 2, "a": 1, "b": 2} == dict({1: 2}, a=1, b=2)  # type:ignore[dict-item]
        )
        assert AttrDict({1: 0}, a=1, b=2, **{"c": 3}) == {1: 0, "a": 1, "b": 2, "c": 3}

    def test_origin_attrs(self):
        d = AttrDict()
        attrs = [i for i in dir(dict) if not i.startswith("__")]
        for index, attr in enumerate(attrs):
            assert callable(getattr(d, attr))
            d[attr] = index
            assert d[attr] == index
            assert callable(getattr(d, attr))

    def test_string_keys_that_can_not_be_attribution(self):
        d = AttrDict({"a-b": 1, ("a", "b"): 2, None: 3, "c_d": 4, "e f": 5})
        assert d["a-b"] == 1
        assert d[("a", "b")] == 2
        with pytest.raises(AttributeError):
            assert d.a_b
        with pytest.raises(AttributeError):
            assert d.ab
        assert d["c_d"] == d.c_d == 4
        assert d["e f"] == 5
        with pytest.raises(AttributeError):
            assert d.e_f
        with pytest.raises(AttributeError):
            assert d.ef

    def test_exclude(self):
        d = AttrDict({"keys": {"a": 1}})
        assert not isinstance(d.keys, AttrDict)
        assert not isinstance(d["keys"], AttrDict)
        assert d.keys() == {"keys": ""}.keys()
        d2 = AttrDict({"_a": {"a": 1}})
        assert not isinstance(d2["_a"], AttrDict)


def test_get_ip(mocker):
    my_ip = get_machine_ip()
    assert re.search(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", my_ip)
    nets = my_ip.split(".")
    assert len(nets) == 4
    assert all(0 <= int(i) <= 255 for i in nets)
    mocker.patch("socket.socket.getsockname", return_value=True)
    assert get_machine_ip() == "127.0.0.1"


def test_cache_attr():
    from asynctor import cache_attr as _cache_attr

    class A:
        def __init__(self) -> None:
            self.started = self.first_instance_created_at()

        @classmethod
        @cache_attr
        def first_instance_created_at(cls) -> datetime:
            return datetime.now()

    a = A()
    b = A()
    now = datetime.now()
    assert a.started == b.started
    c = A()
    assert c.started < now
    assert _cache_attr is cache_attr
    assert a.first_instance_created_at() == b.started


def test_shell():
    from asynctor import Shell as _Shell

    assert _Shell is Shell
    s = Shell("not-exist-command-1")
    with pytest.raises(FileNotFoundError):
        s.call()
    with pytest.raises(FileNotFoundError):
        s.run()
    with pytest.raises(FileNotFoundError):
        s.capture_output()
    assert Shell(["python", "-V"]).call() == 0
    cmd = 'python -c "import os;[print(i) for i in sorted(os.listdir()) if os.path.isfile(i)]"'
    assert Shell(cmd).call() == 0
    assert Shell(cmd, shell=True, capture_output=True).call() == 0
    assert Shell(cmd + "|grep 1111111").call() == 1
    cmd_grep_md = cmd + "|grep md"
    assert Shell(cmd_grep_md, shell=True).call() == 0
    command = shlex.split(cmd) + ["|", "grep", "md"]
    assert Shell(command).command == cmd.replace('"', "'") + " | grep md"
    assert Shell(command, capture_output=True).call() == 0
    assert Shell(command, shell=False, capture_output=True).call() == 0
    out = Shell(command).capture_output()
    out_shell_false = Shell(command, shell=False).capture_output()
    assert out != out_shell_false
    assert out_shell_false == Shell(cmd).capture_output()
    assert out == Shell(cmd_grep_md).capture_output()
    r = Shell.run_by_subprocess(cmd_grep_md, text=True, capture_output=True, encoding="utf-8")
    assert r.stdout == out


def test_shell_redirect(tmp_work_dir):
    file = Path("a")
    file.touch()
    cmd = 'python -c "import os;[print(i) for i in os.listdir() if len(i) < 3]"'
    shell = Shell(cmd)
    assert file.name in shell.capture_output()
    out_file = Path("out.txt")
    shell_out = Shell(f"{cmd} > {out_file}")
    assert file.name not in shell_out.capture_output()
    assert file.name in out_file.read_text()



from __future__ import annotations

import pytest

from asynctor import AsyncTestClient
from asynctor.utils import get_machine_ip

from .main import app


@pytest.fixture(scope="session")
async def client():
    async with AsyncTestClient(app) as c:
        yield c


@pytest.mark.anyio
async def test_client_ip_dep(client):
    path = "/ip"
    r = await client.get(path)
    assert r.status_code == 200
    assert r.json()["client ip"] == "127.0.0.1"

    ip = get_machine_ip()
    keys = "x_forwarded_for", "x_real_ip", "forwarded"
    client.headers[keys[0]] = ip
    r = await client.get(path)
    assert r.status_code == 200
    assert r.json() == {"client ip": ip, "server ip": ip}

    fake_ip = "127.0.0.2"
    client.headers.pop(keys[0])
    client.headers[keys[1]] = fake_ip
    r = await client.get(path)
    assert r.status_code == 200
    assert r.json() == {"client ip": fake_ip, "server ip": ip}

    client.headers.pop(keys[1])
    client.headers[keys[2]] = fake_ip
    r = await client.get(path)
    assert r.status_code == 200
    assert r.json() == {"client ip": "127.0.0.1", "server ip": ip}

    client.headers[keys[2]] = "for=192.0.2.60;proto=http;by=203.0.113.43"
    r = await client.get(path)
    assert r.status_code == 200
    assert r.json() == {"client ip": "192.0.2.60", "server ip": ip}

    client.headers[keys[1]] = fake_ip
    r = await client.get(path)
    assert r.status_code == 200
    assert r.json() == {"client ip": fake_ip, "server ip": ip}

    client.headers[keys[0]] = ip
    r = await client.get(path)
    assert r.status_code == 200
    assert r.json() == {"client ip": ip, "server ip": ip}

    client.headers[keys[0]] = ip + ",192.0.2.60"
    r = await client.get(path)
    assert r.status_code == 200
    assert r.json() == {"client ip": ip, "server ip": ip}



from __future__ import annotations

import functools
import itertools
import sys
import warnings
from collections.abc import AsyncGenerator, Awaitable, Generator, Iterable, Sequence
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar, cast, overload

import anyio
import sniffio
from anyio import from_thread
from anyio._backends._asyncio import get_running_loop
from anyio.lowlevel import checkpoint

from .exceptions import ParamsError

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import ParamSpec, Self, TypeVarTuple, Unpack
else:
    from exceptiongroup import ExceptionGroup  # pragma: no cover
    from typing_extensions import ParamSpec, Self, TypeVarTuple, Unpack  # pragma: no cover

if TYPE_CHECKING:
    from anyio.abc._tasks import TaskGroup

    from ._types import TypeAlias

T = TypeVar("T")
T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")
PosArgsT = TypeVarTuple("PosArgsT")
AsyncFunc: TypeAlias = Callable[..., Awaitable[Any]]


@overload
def ensure_afunc(
    coro: Awaitable[T_Retval],
) -> Callable[[], Awaitable[T_Retval]]: ...


@overload
def ensure_afunc(
    coro: Callable[T_ParamSpec, Awaitable[T_Retval]],
) -> Callable[T_ParamSpec, Awaitable[T_Retval]]: ...


def ensure_afunc(
    coro: Awaitable[T_Retval] | Callable[T_ParamSpec, Awaitable[T_Retval]],
) -> Callable[[], Awaitable[T_Retval]] | Callable[T_ParamSpec, Awaitable[T_Retval]]:
    """Wrap coroutine to be async function"""
    if callable(coro):
        return coro

    async def do_await() -> T_Retval:
        return await coro

    return do_await


def run(
    func: (Awaitable[T_Retval] | Callable[[Unpack[PosArgsT]], Awaitable[T_Retval]]),
    *args: Unpack[PosArgsT],
    backend: str = "asyncio",
    backend_options: dict[str, Any] | None = None,
) -> T_Retval:
    """Combine `asyncio.run` and `anyio.run`

    :param func: async function or coroutine.
    :param args: arguments that will pass to `func` if it's a function.
    :param backend: should be 'asyncio' or 'trio'.
    :param backend_options: will pass to `anyio.run`.

    Usage::

    ... code-block:: python3

        from functools import partial

        async def foo(a, b, *, c=3):
            return a, b, c

        result_asyncio_format = run(foo(1, 2, c=0))
        result_anyio_format = run(partial(foo, c=0), 1, 2)
        assert result_asyncio_format == result_anyio_format

    """
    if not callable(func):

        async def do_await() -> T_Retval:
            return await func

        return anyio.run(do_await, backend=backend, backend_options=backend_options)
    return anyio.run(func, *args, backend=backend, backend_options=backend_options)


class LengthFixedList(list[T]):
    def append(self: Self, _: T) -> None:
        raise TypeError(f"{self.__class__.__name__} is fixed-size")


async def map_group(
    func: AsyncFunc, todos: Iterable[Any], results: list[Any] | None = None
) -> None:
    """
    The `map_group` function asynchronously executes a given function for each set of arguments in a
    provided iterable using `anyio.create_task_group()`.

    :param func: The `func` parameter is an asynchronous function that will be applied to
        each item in the `todos` iterable
    :param todos: The `todos` parameter in the `map_group` function represents an iterable
        containing the arguments that will be passed to the `func` function for each task
        that is started in the task group. Each element in `todos` is a set of arguments
        that will be unpacked and passed to the `func`.
    :param results: The `results` parameter in the `map_group` function is a list that
        stores the results of the asynchronous function calls.
        If the `results` parameter is provided with an initial list,
        the results of each function call will be appended to this list.
        If `results` is None or not provided, tasks will run without any return.
    """
    async with anyio.create_task_group() as tg:
        should_append = results is not None and not isinstance(results, LengthFixedList)
        for args in todos:
            if should_append:
                cast(list[Any], results).append(None)
            tg.start_soon(func, *args)


@overload
async def bulk_gather(
    coros: Sequence[Awaitable[T_Retval]] | Generator[Awaitable[T_Retval]],
    batch_size: int = 0,
    wait_last: bool = False,
    raises: Literal[True] = True,
    *,
    limit: int | None = None,
) -> tuple[T_Retval, ...]: ...


@overload
async def bulk_gather(
    coros: Sequence[Awaitable[T_Retval]] | Generator[Awaitable[T_Retval]],
    batch_size: int = 0,
    wait_last: bool = False,
    raises: Literal[False] = False,
    *,
    limit: int | None = None,
) -> tuple[T_Retval | None, ...]: ...


async def bulk_gather(
    coros: Sequence[Awaitable[T_Retval]] | Generator[Awaitable[T_Retval]],
    batch_size: int = 0,
    wait_last: bool = False,
    raises: bool = True,
    *,
    limit: int | None = None,
) -> tuple[T_Retval | None, ...]:
    """
    Similar like `asyncio.gather`, but support the `batch_size` parameter.
    If batch_size is not zero, running tasks will CapacityLimiter({batch_size}).

    :param coros: Coroutines
    :param batch_size: running tasks limit number, set 0 to be unlimit.
    :param wait_last: if True, wait last bulk tasks to complete then start new task group,
        else use anyio.CapacityLimiter to limit task number.
    :param raises: if True, raise Exception when coroutine failed, else return None.
    :param limit: (deprecated) only leave it here to compare with old version.
    """
    results: list[T_Retval | None]
    try:
        total = len(coros)  # type:ignore[arg-type]
    except TypeError:  # if coros is generator
        total = 0
        results = []
    else:
        if total == 0:
            await checkpoint()
            return ()
        results = LengthFixedList([None] * total)

    async def runner(_i: int, _coro: Awaitable[T_Retval]) -> None:
        results[_i] = await _coro

    async def limited_runner(
        _i: int, _coro: Awaitable[T_Retval], _limiter: anyio.CapacityLimiter
    ) -> None:
        async with _limiter:
            results[_i] = await _coro

    try:
        if limit is not None:
            if batch_size:
                if batch_size != limit:
                    raise ParamsError(f"Conflict value with {limit=} & {batch_size=}")
                warnings.warn(
                    "`limit` is deprecated, use `batch_size` only.",
                    DeprecationWarning,
                    stacklevel=2,
                )
            else:
                batch_size = limit
        if batch_size:
            if wait_last:
                if total == 0:
                    todos: list[tuple[int, Awaitable[T_Retval]]] = []
                    for count, args in zip(itertools.count(), enumerate(coros)):
                        todos.append(args)
                        results.append(None)
                        if count % batch_size == 0:
                            await map_group(runner, todos)
                            todos.clear()
                    if todos:
                        await map_group(runner, todos)
                else:
                    for start in range(0, total, batch_size):
                        coros_slice = itertools.islice(coros, start, start + batch_size)
                        args_items = ((start + i, coro) for i, coro in enumerate(coros_slice))
                        await map_group(runner, args_items)
            else:
                limiter = anyio.CapacityLimiter(batch_size)
                todo_args = ((*item, limiter) for item in enumerate(coros))
                if total == 0:
                    await map_group(limited_runner, todo_args, results)
                else:
                    await map_group(limited_runner, todo_args)
        else:
            if total == 0:
                await map_group(runner, enumerate(coros), results)
            else:
                await map_group(runner, enumerate(coros))
    except ExceptionGroup as e:
        if raises:
            raise e.exceptions[0] from e

    return tuple(results)


async def gather(*coros: Awaitable[T_Retval], limit: int | None = None) -> tuple[T_Retval, ...]:
    """Similar like asyncio.gather, but support set `limit` to control concurrency number.

    :param coros: Coroutines to be executed in the running loop
    :param limit: How many coroutines will be run in the same time (none or zero is unlimit).
    """
    return await bulk_gather(coros, limit=limit)


def create_task(coro: Awaitable[Any], task_group: TaskGroup, *, name: object = None) -> None:
    """Start an async task, similar to asyncio.create_task, but need a task_group param.

    :param coro: Coroutine that will be run in the backgroup
    :param task_group: The backgroup task group

    Usage::
        >>> async def afunc(): ...
        >>> async with anyio.create_task_group() as tg:
        ...     create_task(afunc(), tg)
    """
    task_group.start_soon(ensure_afunc(coro), name=name)


@asynccontextmanager
async def start_tasks(
    coro: Awaitable[Any] | Callable[..., Awaitable[Any]],
    *more: Awaitable[Any] | Callable[..., Awaitable[Any]],
) -> AsyncGenerator[None]:
    """Make it easy to convert asyncio.create_task

    Usage:

    ... code-block:: python3

        async def startup():
            # cost a long time to do sth async
            await anyio.sleep(1000)

        @contextlib.asynccontextmanager
        async def lifespan(app):
            async with start_tasks(startup()):
                yield
    """
    async with anyio.create_task_group() as tg:
        with anyio.CancelScope(shield=True):
            tg.start_soon(ensure_afunc(coro))
            for c in more:
                tg.start_soon(ensure_afunc(c))
            try:
                yield
            finally:
                tg.cancel_scope.cancel()


async def wait_for(coro: Awaitable[T_Retval], timeout: int | float) -> T_Retval:
    """Similar like asyncio.wait_for"""
    with anyio.fail_after(timeout):
        return await coro


def run_async(
    async_func: Awaitable[T_Retval] | Callable[[Unpack[PosArgsT]], Awaitable[T_Retval]],
    *args: Unpack[PosArgsT],
) -> T_Retval:
    """Run async function in worker thread and get the result of it"""
    # `asyncio.run(async_func())`/`anyio.run(async_func)` can get the result of async function,
    # but both of them will close the running loop.
    result: list[T_Retval] = []

    async def runner() -> None:
        if callable(async_func):
            coro = async_func(*args)
            try:
                res = await coro
            except TypeError as e:
                msg = (
                    "object can't be awaited"
                    if sys.version_info >= (3, 14)
                    else "can't be used in 'await' expression"
                )
                if msg in str(e):
                    # In case of async_func is a sync function
                    res = cast(T_Retval, coro)
                    await checkpoint()
                else:
                    raise
        else:
            res = await async_func
        result.append(res)

    with from_thread.start_blocking_portal() as portal:
        portal.call(runner)

    return result[0]


def async_to_sync(
    func: Callable[T_ParamSpec, Awaitable[T_Retval]],
) -> Callable[T_ParamSpec, T_Retval]:
    """Run async function to be sync.

    If there is a running loop, try to run in it,
    otherwise run in worker thread with new loop

    Usage::
        >>> async def do_sth():
        ...     await anyio.sleep(0.1)
        ...     return 1
        >>> res = async_to_sync(do_sth)()
        >>> assert res == 1
    """

    @functools.wraps(func)
    def runner(*args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs) -> T_Retval:
        try:
            asynclib_name = sniffio.current_async_library()
        except sniffio.AsyncLibraryNotFoundError:
            pass
        else:
            if asynclib_name == "asyncio":
                try:
                    loop = get_running_loop()
                except RuntimeError:
                    pass
                else:
                    coro = func(*args, **kwargs)
                    try:
                        return loop.run_until_complete(coro)
                    except RuntimeError as e:
                        if "This event loop is already running" in str(e):
                            return run_async(ensure_afunc(coro))
                        else:
                            raise e
        return run_async(functools.partial(func, **kwargs), *args)

    return runner


def run_until_complete(
    async_func: Awaitable[T_Retval] | Callable[[], Awaitable[T_Retval]],
) -> T_Retval:
    """Run async function or coroutine in runing loop or worker thread"""
    return async_to_sync(ensure_afunc(async_func))()



from __future__ import annotations

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Annotated, Any

from fastapi import Depends, FastAPI, Request
from fastapi.routing import _merge_lifespan_context

from ..client import AsyncRedis


def register_aioredis(
    app: FastAPI,
    check_connection: bool = True,
    **kwargs: Annotated[Any, "Kwargs that will pass to `redis.asyncio.Redis.__init__`"],
) -> None:
    """
    Register redis to fastapi application

    :param app: the fastapi application instance
    :param check_connection: whether check redis server pingable
    :param kwargs: such like: host, port, etc.
    """

    @asynccontextmanager
    async def redis_lifespan(app_instance: FastAPI) -> AsyncGenerator[None]:
        async with AsyncRedis(app_instance, **kwargs):
            yield

    original_lifespan = app.router.lifespan_context
    app.router.lifespan_context = _merge_lifespan_context(redis_lifespan, original_lifespan)


async def get_redis_client(request: Request) -> AsyncRedis:
    return AsyncRedis(request)


AioRedis = Annotated[AsyncRedis, Depends(get_redis_client)]
AioRedisDep = AioRedis
AioRedisDep.__doc__ = """Get register redis cli from application.

Example::

    from asynctor.contrib.fastapi import AioRedisDep, register_aioredis
    from fastapi import FastAPI

    app = FastAPI()
    register_aioredis(app)

    @app.get('/')
    async def get_redis_keys(redis: AioRedisDep) -> list[str]:
        keys = await redis.keys()
        return keys

"""


def get_client_ip(request: Request) -> str:
    if x_forwarded_for := request.headers.get("x_forwarded_for"):
        # X-Forwarded-For may includes many IP address, the first one is origin IP
        return x_forwarded_for.split(",")[0]
    elif x_real_ip := request.headers.get("x_real_ip"):
        return x_real_ip
    elif forwarded := request.headers.get("forwarded"):
        # Forwarded: for=192.0.2.60;proto=http;by=203.0.113.43
        parts = forwarded.split(";")
        for part in parts:
            if part.startswith("for="):
                return part[4:]
    if request.client is None:  # pragma: no cover
        # request.client is not None if server started by uvicorn
        # put it here to improve type hints
        return ""
    return request.client.host


ClientIpDep = Annotated[str, Depends(get_client_ip)]
ClientIpDep.__doc__ = """Get real ip of request client.

Usage::
    >>> @app.get('/')
    >>> def index(client_ip: ClientIpDep):
    ...     assert isinstance(client_ip, str)
"""



from __future__ import annotations

import functools
import inspect
import time
from collections.abc import Awaitable, Callable
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from datetime import datetime, timedelta, timezone
from types import TracebackType
from typing import TYPE_CHECKING, Annotated, Any, TypeVar, overload

try:
    from datetime import UTC  # type:ignore[attr-defined]
except ImportError:
    UTC = timezone.utc

if TYPE_CHECKING:
    from ._types import TypeAlias
    from .compat import Self

T_Retval = TypeVar("T_Retval", Awaitable[Any], Any)
AwareDateTime: TypeAlias = Annotated[datetime, "datetime with tzinfo"]


class Timer(AbstractContextManager, AbstractAsyncContextManager):
    """Print time cost of the function.

    Usage::

    ... code-block:: python3

        import time
        from loguru import logger

        def do_staff():
            time.sleep(0.1)

        with Timer('do sth ...'):
            do_staff()
        # do sth ... Cost: 0.1 seconds

        with Timer('do sth ...', verbose=False) as t:
            do_staff()
        assert isinstance(t.cost, float)
        logger.info(str(t))
        # do sth ... Cost: 0.1 seconds

        @Timer
        async def main():
            do_staff()

        await main()
        # main Cost: 0.1 seconds

        @Timer
        def read_text(filename):
            from pathlib import Path
            time.sleep(0.2222)
            return Path(filename).read_text()

        read_text('a.txt')
        # read_text Cost: 0.2 seconds
    """

    def __init__(self, message: str | Callable, decimal_places=1, verbose=True) -> None:
        if callable(message):  # Use as decorator
            func = message
            self.__name__ = message = func.__name__
            self.func: Callable = func
        self.message = message
        self._decimal_places = decimal_places
        self._end = self._start = time.time()
        self._verbose = verbose

    def start(self) -> None:
        self._start = time.time()

    def capture(self, verbose=None) -> None:
        self._end = time.time()
        if verbose is None:
            verbose = self._verbose
        if verbose:
            print(self)

    @property
    def cost(self) -> float:
        return round(self._end - self._start, self._decimal_places)

    def __str__(self) -> str:
        return f"{self.message} Cost: {self.cost} seconds"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.message!r}, {self._decimal_places}, {self._verbose})"

    async def __aenter__(self) -> Timer:
        return self.__enter__()

    async def __aexit__(self, *args, **kwargs) -> None:
        self.__exit__(*args, **kwargs)

    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if exc_type is not SystemExit or not str(exc_val):
            self.capture()

    def _recreate_cm(self) -> Self:
        return self.__class__(
            getattr(self, "func", None) or self.message,
            self._decimal_places,
            self._verbose,
        )

    def __call__(self, *args, **kwargs) -> Any:
        if (func := getattr(self, "func", None)) is None:
            return None
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def inner(*gs, **kw):
                async with self._recreate_cm():
                    return await func(*gs, **kw)

            return inner(*args, **kwargs)
        else:
            with self._recreate_cm():
                return func(*args, **kwargs)

    @staticmethod
    def now() -> AwareDateTime:
        """Get utc now with tzinfo

        Usage::
            >>> now = Timer.now()
            >>> str(now).endswith('+00:00')
            True
        """
        return datetime.now(UTC)

    @staticmethod
    def to_beijing(dt: AwareDateTime) -> AwareDateTime:
        """Convert tzinfo of aware datetime to beijing timezone"""
        return dt.astimezone(timezone(timedelta(hours=8)))

    @classmethod
    def beijing_now(cls) -> AwareDateTime:
        """Get current beijing time with tzinfo

        Usage::
            >>> now = Timer.beijing_now()
            >>> str(now).endswith('+08:00')
            True
        """
        return cls.to_beijing(cls.now())


@overload
def timeit(func: str) -> Timer: ...  # pragma: no cover


@overload
def timeit(
    func: Callable[..., T_Retval],
) -> Callable[..., T_Retval]: ...  # pragma: no cover


def timeit(func: str | Callable[..., T_Retval]) -> Timer | Callable[..., T_Retval]:
    """Print time cost of the function.

    Usage::

    ... code-block:: python3

        import anyio, time

        @timeit
        async def main():
            await anyio.sleep(0.1)

        await main()
        # main Cost: 0.1 seconds

        @timeit
        def read_text(filename):
            from pathlib import Path
            time.sleep(0.25)
            return Path(filename).read_text()

        read_text('a.txt')
        # read_text Cost: 0.2 seconds

        with timeit('message'):
            time.sleep(0.1)
            await anyio.sleep(1)
        # message Cost: 1.1 seconds

        args, kwargs = (), {}
        def sync_func(): time.sleep(0.24)
        res = timeit(sync_func)(*args, **kwargs)
        # sync_func Cost: 0.2 seconds

        async def async_func():
            await anyio.sleep(1)
        result = await timeit(async_func)(*args, **kwargs)
        # async_func Cost: 1.0 seconds

    """
    if isinstance(func, str):
        return Timer(func)
    func_name = getattr(func, "__name__", str(func))
    if inspect.iscoroutinefunction(func):

        @functools.wraps(func)
        async def deco(*args, **kwargs) -> T_Retval:
            async with Timer(func_name):
                return await func(*args, **kwargs)

    else:

        @functools.wraps(func)
        def deco(*args, **kwargs) -> T_Retval:
            with Timer(func_name):
                return func(*args, **kwargs)

    return deco



from __future__ import annotations

import os

import pytest
from redis.asyncio import Redis

from asynctor import AsyncRedis, AsyncTestClient

from .main import app


@pytest.fixture(scope="session")
async def client():
    async with AsyncTestClient(app) as c:
        yield c


@pytest.mark.anyio
async def test_apis(client):
    path = "/"
    r = await client.get(path)
    assert r.status_code == 200
    assert isinstance(r.json(), list)
    path = "/redis"
    r = await client.get(path, params={"key": "a"})
    assert r.status_code == 200
    v = await AsyncRedis().get("a")
    assert r.json().get("a") == (v and v.decode()), r.text
    payload = {"key": "b", "value": "1"}
    r = await client.post(path, json=payload)
    assert r.status_code == 200, r.text
    assert r.json()["b"] == "1"
    os.environ["REDIS_HOST"] = "localhost"
    cached = await AsyncRedis().get("b")
    assert cached == b"1"


def test_redis_host_port():
    remote_ip = "192.168.0.2"
    redis = AsyncRedis(remote_ip)
    assert isinstance(redis, Redis)
    assert redis.connection_pool.connection_kwargs["host"] == remote_ip

    redis = AsyncRedis(host=remote_ip)
    assert redis.connection_pool.connection_kwargs["host"] == remote_ip

    redis = AsyncRedis("localhost", host=remote_ip)
    assert redis.connection_pool.connection_kwargs["host"] == remote_ip

    custom_port = 8888
    redis = AsyncRedis(port=custom_port)
    assert redis.connection_pool.connection_kwargs["port"] == custom_port


@pytest.mark.anyio
async def test_redis_not_installed(monkeypatch):
    import sys

    monkeypatch.setitem(sys.modules, "redis", None)
    monkeypatch.delattr(Redis, "aclose")

    with pytest.raises(RuntimeError, match=r"pip install redis"):
        async with AsyncRedis():
            pass
    with pytest.raises(RuntimeError, match=r'pip install "asynctor\[redis\]"'):
        async with AsyncRedis(check_connection=False):
            pass



from pathlib import Path

import anyio
import pytest

from asynctor.xls import df_to_datas, load_xls, read_excel


@pytest.mark.anyio
async def test_read():
    demo = Path(__file__).parent / "demo.xlsx"
    df = await read_excel(demo)
    df2 = await read_excel(demo.read_bytes())
    df3 = await read_excel(anyio.Path(demo))
    assert df2.compare(df).empty and df3.compare(df).empty
    data = df_to_datas(df)
    assert data == [
        {"Column1": "row1-\\t%c", "Column2\nMultiLines": 0, "Column 3": 1, 4: ""},
        {"Column1": "r2c1\n00", "Column2\nMultiLines": "r2 c2", "Column 3": 2, 4: ""},
    ]
    assert data == (await load_xls(demo))

    assert (await load_xls(demo, True)) == [
        {"Column1": "row1-\\t%c", "Column2\nMultiLines": "0", "Column 3": "1", 4: ""},
        {"Column1": "r2c1\n00", "Column2\nMultiLines": "r2 c2", "Column 3": "2", 4: ""},
    ]



import subprocess
from pathlib import Path

from asynctor import __version__


def test_version():
    r = subprocess.run(["pdm", "list", "--fields=name,version", "--csv"], capture_output=True)
    out = r.stdout.decode().strip()
    try:
        me = [j for i in out.splitlines() if (j := i.split(","))[0] == "asynctor"][0]
    except IndexError:
        # TODO: remove this when deps of python3.14 can be install by pdm in ci
        assert __version__ in Path("asynctor/__init__.py").read_text()
    else:
        assert me[-1] == __version__



from __future__ import annotations

from enum import auto
from pathlib import Path
from typing import get_type_hints

import pytest

from asynctor.compat import Self, StrEnum, load_toml


class PeopleEnum(StrEnum):
    man = "男"
    female = "女"
    other = "其他"


class AutoEnum(StrEnum):
    First = auto()
    second = auto()
    THIRD = auto()
    fOur = auto()
    five_ = auto()
    _six = auto()


def test_chdir(tmp_work_dir):
    assert tmp_work_dir == Path.cwd()


def test_strenum():
    assert PeopleEnum.man == "男"
    assert PeopleEnum.man is not "男"
    assert PeopleEnum.man.value == "男"
    assert list(PeopleEnum) == [PeopleEnum.man, PeopleEnum.female, PeopleEnum.other]
    assert str(PeopleEnum.man) == "男"
    assert repr(PeopleEnum.man) == "<PeopleEnum.man: '男'>"
    assert [i.value for i in AutoEnum] == ["first", "second", "third", "four", "five_", "_six"]
    with pytest.raises(AttributeError):
        AutoEnum.__seven
    with pytest.raises(TypeError, match="1 is not a string"):

        class A(StrEnum):
            a = 1

    with pytest.raises(TypeError):

        class B(StrEnum):
            a = "a", 1

    with pytest.raises(TypeError):

        class C(StrEnum):
            a = "a", "b", 1

    with pytest.raises(TypeError):

        class D(StrEnum):
            a = "a", "b", "c", "d"

    class E(StrEnum):
        a = b"a", "utf-8"

    assert E.a == "a"

    class F(StrEnum):
        a = "哈喽".encode(), "ascii", "ignore"
        b = "哈喽".encode(), "gbk", "ignore"
        c = "哈喽".encode("gbk"), "gbk", "ignore"

    assert F.a == ""
    assert F.b == "鍝堝柦"
    assert F.c == "哈喽"


def test_load_toml():
    file = Path(__file__).parent.parent / "pyproject.toml"
    content = file.read_text("utf-8")
    assert load_toml(content)["project"]["name"] == "asynctor"


def test_self():
    class A:
        def a_self(self) -> Self:  # Got right type annotation for sub class
            return self

        def a_class(self) -> A:  # Got error type annotation for sub class
            return self

    class B(A):
        _b: str = "B"

    self_b = B().a_self()._b
    class_b = B().a_class()._b  # type:ignore[attr-defined]
    assert self_b == class_b == "B"

    assert get_type_hints(B().a_self)["return"] is Self
    with pytest.raises(NameError):
        (get_type_hints(B().a_class))
